from typing import List, Optional, Union

from Interpreter.Repo import Repo
from Interpreter.Utils import Utils
from Models.AbstractSyntaxTree.AbstractSyntaxTree import LibraryDefinitionNode, ClassDefinitionNode, AbstractSyntaxTree, \
    FunctionDefinitionNode, VariableDefinitionNode, CallDefinitionNode, ValueNode, ASTElement
from Models.Interpreter.Datatypes import Datatypes
from Models.Interpreter.Error import Error, ErrorCodes
from Models.Token.Token import Token
from Models.Token.TokenTypes import TokenTypes
from Models.Token.TokenValueTypes import Keywords, FormatCharacters, ExpressionCharacters


class Parser:

    def __init__(self, repo: Repo, utils: Utils) -> None:
        self.__repo = repo
        self.__utils = utils

        self.__ast: AbstractSyntaxTree = AbstractSyntaxTree()
        self.__saved_tokens: List[Token] = []
        self.__expected_tokens: List[Token] = []

        # for validation if type is created
        self.__is_saving_type = False
        self.__is_saving_value = False
        self.__is_saving_call = False
        self.__is_end = False

        self.__saved_ast_elements: List[ASTElement] = []

    def parse(self, tokens: List[Token]) -> AbstractSyntaxTree:
        self.__is_end = False

        if len(tokens) > 0:
            toks = []
            for tok in tokens:
                toks.append({tok.type.name, tok.value})

            print(self.__repo.line_number, toks)

        for i in range(0, len(tokens)):
            token = tokens[i]

            self.__check_for_expected_tokens(token)

            # keywords
            if token.type == TokenTypes.Keyword:
                if token.value == Keywords.Public.value:
                    self.__saved_tokens.append(token)
                    print('save public')
                    self.__add_expected_token(TokenTypes.Keyword, Keywords.Library.value)
                    self.__add_expected_token(TokenTypes.Keyword, Keywords.Class.value)
                    self.__add_expected_token(TokenTypes.Keyword, Keywords.Function.value)
                    self.__add_expected_token(TokenTypes.Keyword, Keywords.Variable.value)

                elif token.value == Keywords.Library.value:
                    self.__saved_tokens.append(token)
                    print('save lib')
                    self.__add_expected_token(TokenTypes.Name)

                elif token.value == Keywords.Class.value:
                    self.__saved_tokens.append(token)
                    print('save class')
                    self.__add_expected_token(TokenTypes.Name)

                elif token.value == Keywords.Function.value:
                    self.__saved_tokens.append(token)
                    print('save function')
                    self.__add_expected_token(TokenTypes.Name)

                elif token.value == Keywords.Variable.value:
                    self.__saved_tokens.append(token)
                    print('save var')
                    self.__add_expected_token(TokenTypes.Name)

            # names
            elif token.type == TokenTypes.Name:
                if self.__is_saving_variable() and self.__is_saving_value:
                    self.__save_name(token)
                    if self.__is_saving_variable():
                        self.__is_saving_call = True

                # names could be variable types, validation check in evaluator
                elif self.__is_saving_variable() and self.__is_saving_type:
                    self.__save_type(token)
                    self.__add_expected_token(TokenTypes.Format_Character, FormatCharacters.Left_Parenthesis.value)

                elif self.__is_saving_variable():
                    self.__save_name(token)
                    self.__add_expected_token(TokenTypes.Format_Character, FormatCharacters.Colon.value)

                elif self.__is_saving_function():
                    self.__save_name(token)
                    self.__add_expected_token(TokenTypes.Format_Character, FormatCharacters.Left_Parenthesis.value)

                elif self.__is_saving_library() or self.__is_saving_class():
                    self.__save_name(token)
                    self.__add_expected_token(TokenTypes.Format_Character, FormatCharacters.Left_Brace.value)

            # format chars
            elif token.type == TokenTypes.Format_Character:
                if token.value == FormatCharacters.Left_Brace.value:
                    if self.__is_saving_library():
                        self.__save_library()

                    elif self.__is_saving_class():
                        self.__save_class()

                elif token.value == FormatCharacters.Colon.value:
                    if self.__is_saving_variable():
                        self.__is_saving_type = True
                        self.__add_expected_token(TokenTypes.Type)
                        self.__add_expected_token(TokenTypes.Name)

                elif token.value == FormatCharacters.Semicolon.value:
                    self.__end_line()

                elif token.value == FormatCharacters.Left_Parenthesis.value:
                    self.__add_expected_token(TokenTypes.Name)
                    self.__add_expected_token(TokenTypes.Format_Character, FormatCharacters.Right_Parenthesis.value)
                    self.__add_expected_token(TokenTypes.Keyword, Keywords.This.value)
                    self.__add_expected_token(TokenTypes.Name)
                    self.__add_expected_token(TokenTypes.Empty)
                    self.__add_expected_token(TokenTypes.Number)
                    self.__add_expected_token(TokenTypes.String)
                    self.__add_expected_token(TokenTypes.Bool)

                    if self.__is_saving_call:
                        self.__saved_tokens.append(token)

                elif token.value == FormatCharacters.Right_Parenthesis.value:
                    self.__add_expected_token(TokenTypes.Format_Character, FormatCharacters.Semicolon.value)
                    self.__add_expected_token(TokenTypes.Format_Character, FormatCharacters.Colon.value)

                    if self.__is_saving_call:
                        self.__saved_tokens.append(token)
                        self.__save_call()
                        self.__is_saving_call = False

            # expr chars
            elif token.type == TokenTypes.Expression_Character:
                if token.value == ExpressionCharacters.Equal.value:
                    self.__is_saving_value = True
                    self.__add_expected_token(TokenTypes.Bool)
                    self.__add_expected_token(TokenTypes.Number)
                    self.__add_expected_token(TokenTypes.String)
                    self.__add_expected_token(TokenTypes.Empty)
                    self.__add_expected_token(TokenTypes.Name)
                    self.__add_expected_token(TokenTypes.Keyword, Keywords.This.value)
                    self.__add_expected_token(TokenTypes.Keyword, Keywords.Input.value)
                    self.__add_expected_token(TokenTypes.Keyword, Keywords.Range.value)
                    self.__add_expected_token(TokenTypes.Keyword, Keywords.Length.value)
                    self.__add_expected_token(TokenTypes.Format_Character, FormatCharacters.Left_Parenthesis.value)
                    self.__add_expected_token(TokenTypes.Type, Datatypes.Empty.value)

            # types
            elif token.type == TokenTypes.Type:
                if self.__is_saving_variable() and self.__is_saving_value:
                    self.__save_value(token)

                elif self.__is_saving_variable():
                    self.__save_type(token)

            # values
            elif token.type == TokenTypes.Bool or token.type == TokenTypes.Number or token.type == TokenTypes.String:
                if self.__is_saving_call:
                    self.__saved_tokens.append(token)

                elif self.__is_saving_value:
                    self.__save_value(token)

        return self.__ast

    """
        Utils
    """

    def __save_name(self, token: Token):
        self.__saved_tokens.append(token)
        print('save name')

    def __add_expected_token(self, token_type: TokenTypes, value: str = ''):
        self.__expected_tokens.append(Token(token_type, value))

    def __check_for_expected_tokens(self, token: Token):
        error_token: Optional[Token] = None
        is_error = True

        for expected_token in self.__expected_tokens:
            if self.__is_end or token.type != expected_token.type or expected_token.value != '' and token.value != expected_token.value:
                error_token = token
            else:
                is_error = False

        if error_token is not None and is_error:
            self.__utils.error(Error(ErrorCodes.Unexpected, f'{error_token.type.name} {error_token.value}'))

        self.__expected_tokens = []

    def __end_line(self):
        if self.__is_saving_variable():
            self.__save_variable()

        self.__is_end = True
        self.__expected_tokens = []
        self.__is_saving_type = False
        self.__is_saving_value = False

    """
        Library
    """

    def __is_saving_library(self) -> bool:
        found = False
        for token in self.__saved_tokens:
            if token.type == TokenTypes.Keyword and token.value == Keywords.Library.value:
                found = True

        return found

    def __save_library(self) -> None:
        is_public = False
        name_token: Optional[Token] = None

        for token in self.__saved_tokens:
            if token.type == TokenTypes.Name:
                name_token = token
            elif token.type == TokenTypes.Keyword and token.value == Keywords.Public.value:
                is_public = True

        if name_token is not None:
            self.__ast.libraries.append(LibraryDefinitionNode(is_public, name_token.value))
            self.__saved_tokens = []
            self.__add_expected_token(TokenTypes.Keyword, Keywords.Public.value)
            self.__add_expected_token(TokenTypes.Keyword, Keywords.Class.value)
            print('saved library')

    """
        Class
    """

    def __is_saving_class(self) -> bool:
        found = False
        for token in self.__saved_tokens:
            if token.type == TokenTypes.Keyword and token.value == Keywords.Class.value:
                found = True

        return found

    def __save_class(self) -> None:
        is_public = False
        name_token: Optional[Token] = None

        for token in self.__saved_tokens:
            if token.type == TokenTypes.Name:
                name_token = token
            elif token.type == TokenTypes.Keyword and token.value == Keywords.Public.value:
                is_public = True

        if name_token is not None:
            self.__ast.libraries[len(self.__ast.libraries) - 1].classes.append(ClassDefinitionNode(is_public, name_token.value))
            self.__saved_tokens = []
            self.__add_expected_token(TokenTypes.Keyword, Keywords.Public.value)
            self.__add_expected_token(TokenTypes.Keyword, Keywords.Variable.value)
            self.__add_expected_token(TokenTypes.Keyword, Keywords.Function.value)
            print('saved class')

    """
        Call
    """
    def __save_call(self) -> None:
        name_token: Optional[Token] = None
        args: List[ValueNode] = []
        remove_tokens = []
        is_call = False

        for i in range(0, len(self.__saved_tokens)):
            token = self.__saved_tokens[i]
            last_token: Optional[Token] = None
            if i-1 > 0:
                last_token = self.__saved_tokens[i-1]

            if token.type == TokenTypes.Name and last_token is not None and last_token.type == TokenTypes.Format_Character and last_token.value == FormatCharacters.Left_Parenthesis.value:
                pass

            if token.type == TokenTypes.Format_Character and token.value == FormatCharacters.Left_Parenthesis.value and last_token is not None and last_token.type == TokenTypes.Name:
                name_token = last_token
                remove_tokens.append(last_token)
                remove_tokens.append(token)
                is_call = True
            elif is_call and token.type == TokenTypes.Format_Character and token.value == FormatCharacters.Right_Parenthesis.value:
                remove_tokens.append(token)
            elif is_call and token.type == TokenTypes.Bool or token.type == TokenTypes.Number or token.type == TokenTypes.String:
                args.append(ValueNode(token.value, Datatypes[token.type.name]))
                remove_tokens.append(token)

        if name_token is not None:
            call = CallDefinitionNode(name_token.value)
            call.args = args
            self.__saved_ast_elements.append(call)

            for token in remove_tokens:
                self.__saved_tokens.remove(token)

            self.__add_expected_token(TokenTypes.Format_Character, FormatCharacters.Semicolon.value)
            print('saved call')

    """
        Function
    """

    def __is_saving_function(self) -> bool:
        found = False
        for token in self.__saved_tokens:
            if token.type == TokenTypes.Keyword and token.value == Keywords.Function.value:
                found = True

        return found

    def __save_function(self) -> None:
        is_public = False
        name_token: Optional[Token] = None
        return_type: Optional[Token] = None

        for token in self.__saved_tokens:
            if token.type == TokenTypes.Name:
                name_token = token
            elif token.type == TokenTypes.Keyword and token.value == Keywords.Public.value:
                is_public = True
            elif token.type == TokenTypes.Type:
                return_type = token

        if name_token is not None and return_type is not None:
            self.__ast.libraries[len(self.__ast.libraries) - 1].classes[len(self.__ast.libraries) - 1].functions.append(
                FunctionDefinitionNode(is_public, name_token.value, Datatypes[return_type.value]))
            self.__saved_tokens = []
            print('saved function')

    """
        Variable
    """

    def __is_saving_variable(self) -> bool:
        found = False
        for token in self.__saved_tokens:
            if token.type == TokenTypes.Keyword and token.value == Keywords.Variable.value:
                found = True

        return found

    def __save_type(self, token: Token) -> None:
        self.__saved_tokens.append(token)
        self.__add_expected_token(TokenTypes.Format_Character, FormatCharacters.Semicolon.value)
        self.__add_expected_token(TokenTypes.Expression_Character, ExpressionCharacters.Equal.value)
        self.__is_saving_type = False
        print('save type')

    def __save_variable(self):
        is_public = False
        name_token: Optional[Token] = None
        datatype: Optional[Token] = None
        value: Optional[Union[str, Token, CallDefinitionNode]] = None

        reset_saved_ast = False

        for token in self.__saved_tokens:
            if token.type == TokenTypes.Name and name_token is None:
                name_token = token
            elif token.type == TokenTypes.Keyword and token.value == Keywords.Public.value:
                is_public = True
            elif token.type == TokenTypes.Type or name_token is not None and token.type == TokenTypes.Name:
                datatype = token

            value = Token(TokenTypes.Empty, TokenTypes.Empty.value)

        for saved_ast_element in self.__saved_ast_elements:
            if isinstance(saved_ast_element, CallDefinitionNode):
                value = saved_ast_element
                reset_saved_ast = True

        if reset_saved_ast:
            self.__saved_ast_elements = []

        if name_token is not None and datatype is not None and value is not None:
            if not isinstance(value, CallDefinitionNode):
                value = value.value

            if datatype.type == TokenTypes.Name:
                variable = VariableDefinitionNode(is_public, name_token.value, datatype.value, value)
            else:
                variable = VariableDefinitionNode(is_public, name_token.value, Datatypes[str(datatype.value).capitalize()], value)

            if len(self.__ast.libraries) > 0:
                lib = self.__ast.libraries[len(self.__ast.libraries) - 1]
                if len(lib.classes) > 0:
                    cl = lib.classes[len(lib.classes) - 1]
                    if len(cl.functions) == 0:
                        cl.variables.append(variable)
                    else:
                        cl.functions[len(cl.functions) - 1].variables.append(variable)

                    self.__saved_tokens = []
                    print('saved variable')

    """
        Value
    """

    def __save_value(self, token: Token):
        self.__saved_tokens.append(token)
        self.__add_expected_token(TokenTypes.Format_Character, FormatCharacters.Left_Parenthesis.value)
        self.__add_expected_token(TokenTypes.Format_Character, FormatCharacters.Semicolon.value)
        self.__is_saving_value = False
        print('save value')
