from parser.abc.ast import AST
from parser.abc.parser_abc import ParserABC
from parser.model.ast_types_enum import ASTTypesEnum

from cc_lang.model.language_definition_classes import FormatCharacters, Keywords
from cpl_core.console.console import Console
from cpl_query.extension.list import List
from lexer.model.token import Token
from lexer.model.token_types import TokenTypes
from pynput.keyboard import Key
from parser.model.parser_state_enum import ParserStateEnum
from runtime.abc.runtime_service_abc import RuntimeServiceABC
from runtime.model.error import Error
from runtime.model.error_codes_enum import ErrorCodesEnum


class ParserService(ParserABC):

    def __init__(self, runtime: RuntimeServiceABC):
        self._runtime = runtime

        self._access_keywords = [
            Keywords.Public.value,
            Keywords.Private.value,
            Keywords.Static.value,
            Keywords.This.value,
        ]

    def _parse_library_or_class(self, tokens: List[Token], cls=False) -> AST:
        """ Handles library declaration

        Args:
            tokens (List[Token]): Tokens from lexer

        AST:
            lib Main {
            <lib, class> <name> <end>
            public lib Main {
            <access> <lib, class> <name> <end>
            public lib Main {}
            <access> <lib, class> <name> <end> <end>

        Returns:
            AST: Library or class AST
        """
        end = None
        ast = List(AST)
        for i in range(0, tokens.count()):
            token: Token = tokens[i]

            # if line contains }
            if token.type == TokenTypes.Format_Character and token.value == FormatCharacters.Right_Brace.value:
                end = self._runtime.line_count
            elif i == tokens.count()-1 and token.type == TokenTypes.Format_Character and token.value == FormatCharacters.Left_Brace.value:
                break
            elif i == tokens.count()-1:
                self._runtime.error(
                    Error(ErrorCodesEnum.Expected, FormatCharacters.Left_Brace.value))

            if i == 0 and token.type == TokenTypes.Keyword and token.value in self._access_keywords:
                ast.append(AST(ASTTypesEnum.Access, token.value,
                           self._runtime.line_count, self._runtime.line_count))

            if i <= 1 and token.type == TokenTypes.Keyword and token.value == Keywords.Library.value:
                ast.append(AST(ASTTypesEnum.Keyword, token.value,
                           self._runtime.line_count, self._runtime.line_count))

            if i >= 1 and token.type == TokenTypes.Name:
                ast.append(AST(ASTTypesEnum.Name, token.value,
                           self._runtime.line_count, self._runtime.line_count))

        return AST(ASTTypesEnum.LibraryDeclaration if not cls else ASTTypesEnum.ClassDeclaration, ast, self._runtime.line_count, end)

    def _parse_variable(self, tokens: List[AST]) -> AST:
        """ Parses variable declarations

        Args:
            tokens (List[Token]): Tokens from lexer

        AST:
            var test: number;
            <var> <name> <type> <end>
            var test: number = 0;
            <var> <name> <type> <value | int, str, bool> <end>
            var test: number = test;
            <var> <name> <type> <name> <end>
            var test: number = TestClass();
            <var> <name> <type> <object-assign> <end>
            private var test: number = 0;
            <access> <var> <name> <type> <end>

        Returns:
            AST: Library or class AST
        """

    def create_ast(self, tokens: List[Token]) -> List[AST]:
        self._ast = List(AST)

        if tokens.where(lambda t: t.type == TokenTypes.Keyword and t.value == Keywords.Library.value).count() > 0:
            self._ast.append(self._parse_library_or_class(tokens))

        elif tokens.where(lambda t: t.type == TokenTypes.Keyword and t.value == Keywords.Class.value).count() > 0:
            self._ast.append(self._parse_library_or_class(tokens, True))

        elif tokens.where(lambda t: t.type == TokenTypes.Keyword and t.value == Keywords.Variable.value).count() > 0:
            self._ast.append(self._parse_variable(tokens, True))

        return self._ast
